"""Show warning messages."""

import sys

def warning(message: str, *information: str) -> None:
    """Show warning message."""
    print(f"[WARNING] {message}", file=sys.stderr)
    for info in information:
        print(f"  {info}", file=sys.stderr)



# type: ignore
"""Test slipbox.py."""

from time import time

import pytest

from .config import Config
from .slipbox import Slipbox, added_notes, modified_notes, deleted_notes
from .utils import check_requirements, insert_file_script

def test_added_notes_pattern(tmp_path, sbox):
    """added_notes must match the input pattern."""
    slipbox = sbox
    slipbox.config.patterns = ('*.md', '*.txt')

    directory = tmp_path/"directory"
    markdown = tmp_path/"input.md"
    txt = tmp_path/"input.txt"
    tex = tmp_path/"ignore.tex"
    directory.mkdir()
    markdown.touch()
    txt.touch()
    tex.touch()

    notes = added_notes(slipbox)
    assert sorted(notes) == [markdown, txt]

def test_added_notes_in_db(tmp_path, sbox):
    """added_notes must not already be in the database."""
    slipbox = sbox

    new = tmp_path/"new.md"
    skip = tmp_path/"skip.md"
    new.touch()
    skip.touch()
    slipbox.conn.executescript(insert_file_script(skip))
    assert added_notes(slipbox) == [new]

def test_added_notes_recursive(tmp_path, sbox):
    """added_notes must find files recursively."""
    slipbox = sbox

    directory = tmp_path/"directory"
    new = directory/"new.md"
    directory.mkdir()
    new.touch()
    assert added_notes(slipbox) == [new]

def test_slipbox_context_manager(tmp_path):
    """Test database timestamp."""
    config = Config(database=tmp_path/"slipbox.db")
    with Slipbox(config) as slipbox:
        assert slipbox.timestamp == 0.0
    with Slipbox(config) as slipbox:
        assert slipbox.timestamp != 0.0

def test_modified_notes(tmp_path, sbox):
    """modified_notes must exclude the following notes:

    - Notes that aren't in the database
    - Notes in the database that haven't been modified
    """
    modified = tmp_path/"modified.md"
    not_modified = tmp_path/"not_modified.md"
    added = tmp_path/"added.md"

    slipbox = sbox

    slipbox.conn.executescript(insert_file_script(modified, not_modified))

    added.touch()
    modified.touch()

    assert modified_notes(slipbox) == [modified]

def test_deleted_notes(tmp_path, sbox):
    """deleted_notes must exclude the following files:

    - Notes not in the database
    - Notes in the database that still exist in the file system
    """
    delete = tmp_path/"delete.md"
    dont_delete = tmp_path/"dont_delete.md"
    added = tmp_path/"not_in_db.md"

    slipbox = sbox

    slipbox.conn.executescript(insert_file_script(delete, dont_delete))

    dont_delete.touch()
    added.touch()

    assert deleted_notes(slipbox) == [delete]

def test_purge(tmp_path, sbox):
    """Input files must be purged from the database."""
    paths = []
    for prefix in "abcd":
        path = tmp_path/f"{prefix}.md"
        path.touch()
        paths.append(path)

    slipbox = sbox
    slipbox.conn.executescript(insert_file_script(*paths))

    slipbox.purge(paths[2:])

    result = slipbox.conn.execute("SELECT filename FROM Files")
    remaining = sorted(filename for filename, in result)
    assert len(remaining) == 2
    for path, filename in zip(paths, remaining):
        assert path.samefile(filename)

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_suggest_edits_backlinks(files_abc, sbox):
    """slipbox.suggest_edits must include backlinks of outdated notes."""
    file_a, file_b, file_c = files_abc
    file_a.write_text("# 0 A\n\nA.\n[B](#1 '/a').\n")
    file_b.write_text("# 1 B\n\nB.\n[C](#2).\n")
    file_c.write_text("# 2 C\n\nC.\n")

    slipbox = sbox
    slipbox.timestamp = time()

    slipbox.process([file_a, file_b, file_c])

    file_c.touch()

    notes = slipbox.find_notes()
    suggestions = list(slipbox.suggest_edits(notes))
    assert suggestions == [('1', "B", file_b)]

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_suggest_edits_aliases(files_abc, sbox):
    """slipbox.suggest_edits must include alias owners of outdated notes."""
    file_a, file_b, file_c = files_abc
    file_a.write_text("# 0 A\n\nA.\n[B](#1 '/a').\n")
    file_b.write_text("# 1 B\n\nB.\n[C](#2).\n")
    file_c.write_text("# 2 C\n\nC.\n")

    slipbox = sbox
    slipbox.timestamp = time()

    slipbox.process([file_a, file_b, file_c])

    file_b.touch()

    notes = slipbox.find_notes()
    suggestions = list(slipbox.suggest_edits(notes))
    assert suggestions == [('0', "A", file_a)]

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_suggest_edits_exclude_deleted_notes(files_abc, sbox):
    """slipbox.suggest_edits must exclude deleted notes."""
    file_a, file_b, file_c = files_abc
    file_a.write_text("# 0 A\n\nA.\n[B](#2 '/a').\n")
    file_b.write_text("# 1 B\n\nB.\n[C](#2).\n")
    file_c.write_text("# 2 C\n\nC.\n")

    slipbox = sbox
    slipbox.timestamp = time()

    slipbox.process([file_a, file_b, file_c])

    file_c.touch()
    file_a.unlink()

    notes = slipbox.find_notes()
    suggestions = list(slipbox.suggest_edits(notes))
    assert suggestions == [('1', "B", file_b)]

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_run(files_abc, capsys, sbox):
    """There must be no suggestions when running for the first time."""
    file_a, file_b, file_c = files_abc
    file_a.write_text("# 0 A\n\nA.\n[B](#2 '/a').\n")
    file_b.write_text("# 1 B\n\nB.\n[C](#2).\n")
    file_c.write_text("# 2 C\n\nC.\n")

    slipbox = sbox
    slipbox.timestamp = time()

    slipbox.run()

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process(tmp_path, sbox):
    """Smoke test for slipbox.process."""
    input_file = tmp_path/"input.md"
    input_file.write_text("# 1 Test note\n\nHello, world!\n")
    assert not list(sbox.conn.execute("SELECT * FROM Html"))

    sbox.process([input_file])
    result = list(sbox.conn.execute("SELECT * FROM Sections"))
    # Check if the output HTML has sections.
    assert len(result) == 1

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_empty_file(tmp_path, sbox):
    """Scanned files that are empty shouldn't have entries in the database."""
    empty = tmp_path/"empty.md"
    empty.touch()
    sbox.process([empty])
    assert not list(sbox.conn.execute("SELECT * FROM Files"))
    assert not list(sbox.conn.execute("SELECT * FROM Notes"))
    assert not list(sbox.conn.execute("SELECT * FROM Tags"))
    assert not list(sbox.conn.execute("SELECT * FROM Links"))
    assert not list(sbox.conn.execute("SELECT * FROM Aliases"))
    assert not list(sbox.conn.execute("SELECT * FROM Sequences"))
    assert not list(sbox.conn.execute("SELECT * FROM Html"))
    assert not list(sbox.conn.execute("SELECT * FROM Sections"))
    assert not list(sbox.conn.execute("SELECT * FROM Bibliography"))
    assert not list(sbox.conn.execute("SELECT * FROM Citations"))

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_filenames(tmp_path, sbox):
    """Filenames must be scanned correctly."""
    markdown = tmp_path/"foo.md"
    skip = tmp_path/"bar.md"
    markdown.write_text("# 0 Note\n\nBody.\n")
    skip.write_text("# 0 Note\n\nBody.\n")
    sbox.process([markdown])

    result = list(sbox.conn.execute("SELECT filename FROM Notes WHERE id = '0'"))
    assert len(result) == 1
    assert markdown.samefile(result[0][0])

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_filenames0(tmp_path, sbox):
    """Filenames must be scanned correctly."""
    markdown = tmp_path/"bar.md"
    skip = tmp_path/"foo.md"
    markdown.write_text("# 0 Note\n\nBody.\n")
    skip.write_text("# 0 Note\n\nBody.\n")
    sbox.process([markdown])
    result = list(sbox.conn.execute("SELECT filename FROM Notes WHERE id = '0'"))
    assert len(result) == 1
    assert markdown.samefile(result[0][0])

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_non_level1_headers(tmp_path, sbox):
    """Only level 1 headers must be considered as note headers."""
    markdown = tmp_path/"test.md"
    markdown.write_text("""# 0 Valid note header

Foo.

## 1 Invalid note header

Bar.
""")
    sbox.process([markdown])
    result = [nid for nid, in sbox.conn.execute("SELECT id FROM Notes")]
    assert len(result) == 1
    assert result == ['0']

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_with_duplicate_existing_id(tmp_path, sbox, capsys):
    """slipbox.process must show a warning if a new note shares the ID of an
    existing note.

    The warning message must show the filenames of both notes.
    """
    file_a = tmp_path/"a.md"
    file_a.write_text("# 0 Existing note\n\nTest.\n")

    sbox.process([file_a])
    result = list(sbox.conn.execute("SELECT id, title FROM Notes"))
    assert len(result) == 1
    assert result == [('0', "Existing note")]

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr

    file_b = tmp_path/"b.md"
    file_b.write_text("# 0 Duplicate note\n\nTest.\n")

    sbox.process([file_b])
    result = list(sbox.conn.execute("SELECT id, title FROM Notes"))
    assert len(result) == 1
    assert result == [('0', "Existing note")]

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr
    assert str(file_a) in stderr
    assert str(file_b) in stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_with_duplicate_ids_in_a_file(tmp_path, capsys, sbox):
    """If there are duplicate IDs in a file, only the first one must be saved.

    slipbox.process must show a warning when this happens.
    """
    markdown = tmp_path/"test.md"
    markdown.write_text("""# 0 First note

Foo.

# 0 Duplicate

Bar.
""")
    sbox.process([markdown])
    result = list(sbox.conn.execute("SELECT id, title FROM Notes"))
    assert len(result) == 1
    assert result == [('0', "First note")]

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_with_missing_alias(tmp_path, capsys, sbox):
    """slipbox.process must show a warning if there's a gap in an alias sequence."""
    markdown = tmp_path/"test.md"
    markdown.write_text("""# 0 Foo

[Bar](#1 '/a')
[Baz](#2 '/a1a')

# 1 Bar

Bar.

# 2 Baz

Baz.
""")
    sbox.process([markdown])
    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_aliases(sbox, tmp_path):
    """Only slash aliases must be saved."""
    file_a = tmp_path/"a.md"
    file_b = tmp_path/"b.md"
    file_c = tmp_path/"c.md"
    file_a.write_text("""# 0 A

[B](#1 '0a')
[C](#2 '/b')
""")
    file_b.write_text("# 1 B\n\nB")
    file_c.write_text("# 2 C\n\nC")

    sbox.process([file_a, file_b, file_c])
    result = list(sbox.conn.execute("SELECT id, alias, owner FROM Aliases ORDER BY alias"))
    assert len(result) == 2
    assert result == [('0', '0', '0'), ('2', '0b', '0')]

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_with_empty_link_target(tmp_path, capsys, sbox):
    """slipbox.process must show a warning if there is a link with an empty target."""
    markdown = tmp_path/"test.md"
    markdown.write_text("# 0 Foo\n\n[Empty]().\n")
    sbox.process([markdown])
    result = list(sbox.conn.execute("SELECT * FROM Links"))
    assert not result

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_with_id_in_scientific_form(tmp_path, capsys, sbox):
    """Headers with non-integer IDs should be ignored."""
    markdown = tmp_path/"test.md"
    markdown.write_text("# 1e1 Invalid note ID\n\nTest.\n")
    sbox.process([markdown])

    result = list(sbox.conn.execute("SELECT * FROM Notes"))
    assert not result

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_with_non_text_titles(tmp_path, capsys, sbox):
    """Notes with non-text titles in the header must still be recognized."""
    file_a = tmp_path/"a.md"
    file_b = tmp_path/"b.md"
    file_c = tmp_path/"c.md"
    file_a.write_text("# 0 $1 + 1 = 2$\n\nTest.\n")
    file_b.write_text("# 1 [Note 0](#0)\n\nTest.\n")
    file_c.write_text("# 2 `print('Title')`\n\nTest.\n")

    sbox.process([file_a, file_b, file_c])
    result = list(sbox.conn.execute("SELECT id, title FROM Notes ORDER BY id"))

    assert len(result) == 3
    assert result == [('0', "1 + 1 = 2"), ('1', "Note 0"), ('2', "print('Title')")]

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert not stderr

@pytest.mark.skipif(not check_requirements(), reason="requires pandoc")
def test_process_with_duplicate_aliases(tmp_path, capsys, sbox):
    """If a note defines duplicate aliases, only the first one must be saved.

    slipbox.process must show a warning when this happens.
    """
    markdown = tmp_path/"test.md"
    markdown.write_text("""# 0 Foo

[Bar](#1 '/a')
[Baz](#2 '/a')

# 1 Bar

Bar.

# 2 Baz

Baz.
""")
    sbox.process([markdown])
    result = list(sbox.conn.execute("SELECT id FROM Aliases WHERE alias = '0a'"))
    assert len(result) == 1
    assert result == [('1',)]

    stdout, stderr = capsys.readouterr()
    assert not stdout
    assert stderr



# type: ignore
"""Functions for mocking the database."""

from pathlib import Path
import sqlite3
from typing import Iterable, List

import pytest

from . import scan
from .config import Config
from .slipbox import Slipbox

@pytest.fixture
def mock_db() -> Iterable[sqlite3.Connection]:
    """Create an empty mock database with all the necessary tables."""
    with sqlite3.connect(":memory:") as conn:
        scan.initialize_database(conn)
        yield conn

@pytest.fixture
def mconfig(tmp_path) -> Config:
    """Create mock config object.

    - config.database is tmp_path/slipbox.db
    - config.paths consists of tmp_path
    """
    yield Config(database=tmp_path/"slipbox.db",
                 paths=(tmp_path,))

@pytest.fixture
def sbox(mconfig) -> Slipbox:
    """Create automatically configured Slipbox object."""
    with Slipbox(mconfig) as slipbox:
        yield slipbox

@pytest.fixture
def files_abc(tmp_path) -> List[Path]:
    """Create files in tmp_path: a.md, b.md, c.md."""
    files = [tmp_path/"a.md", tmp_path/"b.md", tmp_path/"c.md"]
    for path in files:
        path.touch()
    yield files



"""Sqlite database wrapper."""

from collections import namedtuple
from itertools import chain
import os.path
from pathlib import Path
import sqlite3
from typing import Iterable, Iterator, List, Tuple

from . import scan, page
from .config import Config
from .data import warning
from .utils import sqlite_string

Notes = namedtuple("Notes", "added modified deleted")

class Slipbox:
    """Slipbox main functions."""
    def __init__(self, config: Config = Config()):
        self.timestamp = 0.0
        self.config = config
        if config.database.exists():
            self.timestamp = os.path.getmtime(config.database)
        self.conn = sqlite3.connect(config.database)

        sql = Path(__file__).with_name("schema.sql").read_text()
        self.conn.executescript(sql)
        self.conn.commit()

    def close(self) -> None:
        """Close database connection."""
        self.conn.close()

    def __enter__(self) -> "Slipbox":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None: # type: ignore
        self.close()

    def find_notes(self) -> Notes:
        """Return a named tuple containing lists of new, modified and deleted notes."""
        return Notes(
            added=added_notes(self),
            modified=modified_notes(self),
            deleted=deleted_notes(self),
        )

    def suggest_edits(self, notes: Notes) -> Iterator[Tuple[int, str, Path]]:
        """Suggest notes to edit based on the given set of notes.

        The suggestion includes
        - notes that link to notes that have been modified/deleted
        - notes that define an alias for modified/deleted notes
        and excludes notes that have been deleted.
        """
        outdated = chain(notes.modified, notes.deleted)
        filenames = [sqlite_string(str(path)) for path in outdated]

        sql = """
            SELECT owner FROM Notes JOIN Aliases USING (id)
                WHERE filename IN ({})
        """.format(",".join(filenames))
        owners = self.conn.execute(sql)

        sql = """
            SELECT src FROM Notes JOIN Links ON id = dest
                WHERE filename IN ({})
        """.format(",".join(filenames))
        backlinks = self.conn.execute(sql)

        affected = (repr(nid) for nid, in chain(owners, backlinks))

        sql = """
            SELECT id, title, filename FROM Notes
                WHERE id IN ({}) ORDER BY id
        """.format(",".join(affected))

        for nid, title, filename in self.conn.execute(sql):
            path = Path(filename)
            if path not in notes.deleted:
                yield nid, title, path

    def purge(self, paths: Iterable[Path]) -> None:
        """Purge filenames from the database."""
        filenames = ((str(path),) for path in paths)
        cur = self.conn.cursor()
        cur.execute("PRAGMA foreign_keys=ON")
        cur.executemany("DELETE FROM Files WHERE filename IN (?)", filenames)

    def process(self, paths: Iterable[Path]) -> None:
        """Process input files."""
        inputs = list(set(paths))
        for batch in scan.group_by_file_extension(inputs):
            scan.process_batch(self.conn, list(batch), self.config)

    def compile(self) -> None:
        """Compile processed HTML into final output."""
        page.generate_complete_html(self.conn, self.config.document_options)

    def run(self) -> None:
        """Run all steps needed to compile output."""
        notes = self.find_notes()
        suggestions = list(self.suggest_edits(notes))
        self.purge(chain(notes.modified, notes.deleted))
        self.process(chain(notes.added, notes.modified))
        self.compile()
        if suggestions:
            warning(
                "The notes below are related to notes that have recently been updated.",
                "You might want to review them for inconsistent links.",
                *(f"  {nid}. {title} in {str(path)!r}."
                  for nid, title, path in suggestions)
            )

def added_notes(slipbox: Slipbox) -> List[Path]:
    """Return list of newly added notes."""
    paths = slipbox.config.paths
    patterns = slipbox.config.patterns
    added = scan.find_new_files(slipbox.conn, paths, patterns)
    return list(added)

def modified_notes(slipbox: Slipbox) -> List[Path]:
    """Return list of notes modified since the last scan."""
    files = scan.fetch_files(slipbox.conn)
    return [p for p in files if scan.is_recently_modified(slipbox.timestamp, p)]

def deleted_notes(slipbox: Slipbox) -> List[Path]:
    """Return list of notes that have been deleted from the file system."""
    files = scan.fetch_files(slipbox.conn)
    return [p for p in files if not p.exists()]



"""Build HTML from all notes in slipbox."""

from argparse import ArgumentParser
from pathlib import Path
import sys

from .config import Config
from .slipbox import Slipbox
from .utils import check_requirements

def main(config_: Config) -> None:
    """Compile notes into static page."""
    with Slipbox(config_) as slipbox:
        slipbox.run()

if __name__ == "__main__":
    if not check_requirements():
        sys.exit("[ERROR] pandoc not found.")

    config = Config()

    parser = ArgumentParser(
        description="Generate a single-page HTML from your notes.")
    parser.add_argument("-s", "--database", default=config.database, type=Path,
                        help="filename of sqlite3 database")
    parser.add_argument("-P", "--paths", nargs='+', default=config.paths, type=Path,
                        help="list of files or directories that contain notes")
    parser.add_argument("-p", "--patterns", nargs='*', default=config.patterns,
                        help="list of glob patterns")
    parser.add_argument("-c", "--content-options", default=config.content_options,
                        help="pandoc options for the content")
    parser.add_argument("-d", "--document-options", default=config.document_options,
                        help="pandoc options for the final HTML output")
    parser.add_argument("--convert-to-data-url", action="store_true",
                        default=config.convert_to_data_url,
                        help="convert local image links to data URL")
    parser.parse_args(namespace=config)

    main(config)



"""Slipbox configuration."""

from dataclasses import dataclass
from pathlib import Path
from typing import Sequence

_CSS = Path(__file__).with_name("pandoc.css")

@dataclass
class Config:
    """Slipbox config object."""
    database: Path = Path("slipbox.db")
    paths: Sequence[Path] = (Path(),)
    patterns: Sequence[str] = ('*.md',)
    content_options: str = "--mathjax"
    document_options: str = "--mathjax -H {}".format(_CSS.resolve())
    convert_to_data_url: bool = False



# type: ignore
"""Test scan.py."""

import os
import time

import pytest

from . import scan
from .utils import insert_file_script

def test_is_recently_modified(tmp_path):
    """is_recently_modified should return true iff the input file was modified
    after the specified timestamp.
    """
    before = time.time()
    path = tmp_path/"file"
    path.touch()
    os.utime(path, ns=(time.time_ns(), time.time_ns()))
    after = time.time()

    assert scan.is_recently_modified(before, path)
    assert not scan.is_recently_modified(after, path)

    os.utime(path, ns=(time.time_ns(), time.time_ns()))
    assert scan.is_recently_modified(after, path)

def test_is_file_in_db(mock_db, tmp_path):
    """Quick check for is_file_in_db."""
    conn = mock_db
    present = tmp_path/"present"
    absent = tmp_path/"absent"
    conn.executescript(insert_file_script(present.name))
    assert scan.is_file_in_db(present.name, conn)
    assert not scan.is_file_in_db(absent.name, conn)

def test_has_valid_pattern():
    """has_valid_pattern is true only for filenames that match at least one of
    the wildcard patterns."""
    patterns = ("*.md", "*.rst")
    assert scan.has_valid_pattern("a.md", patterns)
    assert scan.has_valid_pattern("a.rst", patterns)
    assert not scan.has_valid_pattern("a.tex", patterns)
    assert not scan.has_valid_pattern("a.md", (".md"))

def test_glob_files(tmp_path):
    """glob_files should return set of files in the given set of paths."""
    directory = tmp_path/"directory"
    directory.mkdir()
    directory.joinpath("subdirectory").mkdir()

    file_a = tmp_path/"a.md"
    file_b = tmp_path/"b.md"
    file_c = tmp_path/"directory"/"c.md"
    file_d = tmp_path/"directory"/"d.md"
    file_e = tmp_path/"directory"/"subdirectory"/"e.md"

    file_a.touch()
    file_b.touch()
    file_c.touch()
    file_d.touch()
    file_e.touch()

    inputs = [file_a, file_b, directory]
    files = scan.glob_files(inputs)
    assert len(files) == 5
    assert file_a in files
    assert file_b in files
    assert file_c in files
    assert file_d in files
    assert file_e in files

def test_find_new_files(mock_db, tmp_path):
    """find_new_files must only return existing files that aren't yet in the
    database and match the input patterns (*.md by default).
    """
    conn = mock_db
    present = tmp_path/"present.md"
    absent = tmp_path/"absent.md"
    directory = tmp_path/"directory"
    txt = tmp_path/"ignore.txt"
    present.touch()
    absent.touch()
    directory.mkdir()
    txt.touch()

    conn.executescript(insert_file_script(present))
    new_files = list(scan.find_new_files(conn, [tmp_path]))
    assert new_files == [absent]

def test_group_by_file_extension():
    """group_by_file_extension should split by file type.

    Files with no extension should be considered their own type.
    """

    files = ["a.md", "b.md", ".md", "c.tex", ".tex", ""]
    groups = list(map(list, scan.group_by_file_extension(files)))
    assert len(groups) == 5
    assert ["a.md", "b.md"] in groups
    assert [".md"] in groups
    assert ["c.tex"] in groups
    assert [".tex"] in groups
    assert [""] in groups

def test_group_by_file_extension_on_the_same_type():
    """group_by_file_extension should group files with the same type together."""
    files = [f"{c}.md" for c in "abcdefg"]
    groups = list(map(list, scan.group_by_file_extension(files)))
    assert len(groups) == 1
    assert files in groups

def test_build_command(tmp_path):
    """Sanity check for build_command."""
    input_file = tmp_path/"input.md"
    output = "output.html"
    options = "--mathjax"
    input_file.touch()

    cmd = scan.build_command(input_file, output, options)
    assert str(input_file) in cmd
    assert f"-o {output}" in cmd
    assert options in cmd

@pytest.mark.xfail
def test_build_command_when_input_file_does_not_exist(tmp_path):
    """build_command must fail if input file does not exist."""
    input_file = tmp_path/"input.md"
    scan.build_command(input_file, "output.html", "")
